package nl.kallestruik.dtweaks.managers

import nl.kallestruik.dtweaks.Const
import nl.kallestruik.dtweaks.tweaks.ITweak
import org.bukkit.configuration.InvalidConfigurationException
import org.bukkit.configuration.file.YamlConfiguration
import org.bukkit.plugin.java.JavaPlugin
import java.io.File
import java.io.IOException
import java.util.concurrent.ConcurrentHashMap

class TweakManager(
    private val plugin: JavaPlugin
) {

    // A HashMap containing the tweak identifier mapped to the instance of the tweak.
    private val tweaks: ConcurrentHashMap<String, ITweak> = ConcurrentHashMap<String, ITweak>()

    // A HashMap containing categories and the tweak identifiers that belong to them.
    private val tweakCategories = HashMap<String, List<String>>()

    // A HashMap containing the sate of all the known tweaks.
    private val tweakStates = HashMap<String, Boolean>()


    /**
     * Register a new tweak.
     *
     * @param tweak          An instance of the tweak.
     * @return True if the tweak did not yet exist false otherwise.
     */
    fun registerTweak(tweak: ITweak): Boolean {
        if (tweaks.containsKey(tweak.getIdentifier())) return false
        tweaks[tweak.getIdentifier()] = tweak
        tweakCategories[tweak.getIdentifier()] = tweak.getCategories()
        tweak.onRegister()
        if (tweakStates.getOrDefault(tweak.getIdentifier(), false)) tweak.onEnable()
        return true
    }

    /**
     * Unregister a tweak.
     *
     * @param identifier The identifier of the tweak to unregister.
     * @return True if the tweak was successfully unregistered false otherwise.
     */
    fun unRegisterTweak(identifier: String): Boolean {
        if (!tweaks.containsKey(identifier)) return false
        val tweak: ITweak? = tweaks[identifier]
        tweak?.onDisable()
        tweak?.onUnRegister()
        tweaks.remove(identifier)
        return true
    }

    /**
     * Unregister every tweak that is currently registered.
     */
    fun unRegisterAllTweaks() {
        for (identifier in tweaks.keys) {
            unRegisterTweak(identifier)
        }
    }

    /**
     * Enable a tweak.
     *
     * @param identifier The identifier of the tweak.
     * @return True if the tweak exists false otherwise.
     */
    fun enableTweak(identifier: String): Boolean {
        if (!tweaks.containsKey(identifier)) return false
        tweakStates[identifier] = true
        tweaks[identifier]?.onEnable()
        return true
    }

    /**
     * Disable a tweak.
     *
     * @param identifier The identifier of the tweak.
     * @return True if the tweak exists false otherwise.
     */
    fun disableTweak(identifier: String): Boolean {
        if (!tweaks.containsKey(identifier)) return false
        tweakStates[identifier] = false
        tweaks[identifier]?.onDisable()
        return true
    }

    @Throws(IOException::class, InvalidConfigurationException::class)
    fun loadFromFile(file: File) {
        if (!file.parentFile.exists()) file.parentFile.mkdirs()
        if (!file.exists())
            plugin.saveResource(file.name, false)

        val config = YamlConfiguration()
        config.load(file)
        for (key in config.getKeys(false)) {
            tweakStates[key] = config.getBoolean(key!!)
        }
    }

    @Throws(IOException::class)
    fun saveToFile(file: File) {
        val config = YamlConfiguration()
        for ((key, value) in tweakStates.entries) {
            config[key] = value
        }
        config.save(file)
    }

    fun isTweakEnabled(tweakId: String): Boolean {
        return tweakStates[tweakId] ?: false
    }
}