package nl.kallestruik.dtweaks.tweaks.miscellaneoustweaks

import com.comphenix.protocol.PacketType
import com.comphenix.protocol.ProtocolLibrary
import com.comphenix.protocol.events.PacketAdapter
import com.comphenix.protocol.events.PacketEvent
import com.comphenix.protocol.wrappers.BlockPosition
import com.comphenix.protocol.wrappers.EnumWrappers.Hand
import nl.kallestruik.dlib.MathHelper
import nl.kallestruik.dtweaks.tweaks.ITweak
import org.bukkit.Bukkit
import org.bukkit.block.BlockFace
import org.bukkit.block.data.Bisected
import org.bukkit.block.data.BlockData
import org.bukkit.block.data.Directional
import org.bukkit.block.data.type.Comparator
import org.bukkit.block.data.type.Repeater
import org.bukkit.block.data.type.Stairs
import org.bukkit.block.data.type.TrapDoor
import org.bukkit.entity.Player
import org.bukkit.event.Listener
import org.bukkit.inventory.ItemStack
import org.bukkit.plugin.java.JavaPlugin

class CarpetBlockPlacingProtocol(
    private val plugin: JavaPlugin,
    private val mathHelper: MathHelper
): ITweak, Listener {
    private val protocolManager = ProtocolLibrary.getProtocolManager()
    private var packetListener: PacketListener? = null

    override fun getIdentifier(): String = "CarpetBlockPlacingProtocol"
    override fun getCategories(): List<String> = listOf("player", "survival")

    override fun onEnable() {
        this.packetListener = PacketListener(plugin, PacketType.Play.Client.USE_ITEM, mathHelper)
        protocolManager.addPacketListener(this.packetListener)
    }

    override fun onDisable() {
        protocolManager.removePacketListener(this.packetListener)
    }

    internal class PacketListener(
        plugin: JavaPlugin?,
      packetType: PacketType?,
      private val mathHelper: MathHelper
      ): PacketAdapter(plugin, packetType) {
        override fun onPacketReceiving(event: PacketEvent) {
            // Make sure the packet is the one we expect.
            if (event.packetType !== PacketType.Play.Client.USE_ITEM) return

            // Get the hand that the player used.
            val hand = event.packet.hands.values[0] ?: return

            // Get the MovingObjectPositionBlock and make sure it is not null.
            val mopd = event.packet.movingBlockPositions.values[0] ?: return
            val player = event.player

            // Get the item stack that was clicked with.
            val itemStack =
                if (hand == Hand.MAIN_HAND) player.inventory.itemInMainHand else player.inventory.itemInOffHand

            // Get the hitPos and BlockPos from the mopd.
            val hitPos = mopd.posVector
            val pos = mopd.blockPosition

            // Calculate hitX.
            val hitX = hitPos.x - pos.x

            // Nothing special so ignore this packet.
            if (hitX < 2) return
            val code = (hitX - 2).toInt() / 2
            val blockData: BlockData = getBlockDataFromCode(code, itemStack, player, pos) ?: return
            Bukkit.getScheduler().runTask(plugin, Runnable {
                val blockToChange = player.world.getBlockAt(pos.x, pos.y, pos.z)
                blockToChange.blockData = blockData

                // Cancel any further processing
                event.isCancelled = true

                // Take one item from the players inventory.
                itemStack.amount = itemStack.amount - 1
            })
        }

        private fun getBlockDataFromCode(code: Int, itemStack: ItemStack, player: Player, pos: BlockPosition): BlockData? {
            val blockData = Bukkit.createBlockData(itemStack.type)
            if (blockData is Repeater) {
                val repeater = blockData
                val delay: Int = mathHelper.clamp(code shr 4, 1, 4)
                repeater.facing = horizontalBlockFaceFromCode(code and 0xF, player)!!
                repeater.delay = delay
                return repeater
            } else if (blockData is TrapDoor) {
                val trapDoor = blockData
                val block = player.world.getBlockAt(pos.x, pos.y, pos.z)
                trapDoor.facing = blockFaceFromCode(code and 0xF)
                trapDoor.half = if (code > 0xF) Bisected.Half.TOP else Bisected.Half.BOTTOM
                trapDoor.isOpen = block.isBlockPowered || block.isBlockIndirectlyPowered
                return trapDoor
            } else if (blockData is Comparator) {
                val comparator = blockData
                val mode = if (code > 0xF) Comparator.Mode.SUBTRACT else Comparator.Mode.COMPARE
                comparator.facing = horizontalBlockFaceFromCode(code and 0xF, player)!!
                comparator.mode = mode
                return comparator
            } else if (blockData is Stairs) {
                val stairs = blockData
                stairs.facing = blockFaceFromCode(code and 0xF)
                stairs.half = if (code > 0xF) Bisected.Half.TOP else Bisected.Half.BOTTOM
                return stairs
            } else if (blockData is Directional) {
                val directional = blockData
                val possibilities = directional.faces

                // Check what type of directional we are dealing with.
                if (possibilities.contains(BlockFace.UP)) {
                    directional.facing = blockFaceFromCode(code)
                } else {
                    directional.facing = horizontalBlockFaceFromCode(code, player)!!
                }
                return directional
            }
            return null
        }

        private fun blockFaceFromCode(code: Int): BlockFace {
            return when (code) {
                0 -> BlockFace.DOWN
                1 -> BlockFace.UP
                2 -> BlockFace.NORTH
                3 -> BlockFace.SOUTH
                4 -> BlockFace.WEST
                5 -> BlockFace.EAST
                else -> BlockFace.SELF
            }
        }

        private fun blockFaceFromRotation(rotation: Float): BlockFace? {
            return if (rotation < 45 && rotation >= -45) BlockFace.SOUTH else if (rotation < 135 && rotation >= 45) BlockFace.WEST else if (rotation >= 135 && rotation < -135) BlockFace.NORTH else BlockFace.EAST
        }

        private fun horizontalBlockFaceFromCode(code: Int, player: Player): BlockFace? {
            val face = blockFaceFromCode(code)
            return if (face == BlockFace.UP || face == BlockFace.DOWN) {
                blockFaceFromRotation(player.location.yaw)
            } else face
        }
    }
}